package dk.nsp.idsas.security.impl;

import java.time.ZonedDateTime;
import java.util.LinkedList;
import java.util.List;
import java.util.function.Supplier;

import dk.nsp.idsas.security.IdsasSecurityException;
import dk.nsp.idsas.security.model.IdsasActor;
import dk.nsp.idsas.util.ErrorMessages;
import dk.sds.nsp.security.Security;
import dk.sds.nsp.security.SecurityContext;

import dk.nsp.idsas.security.IdsasActorProvider;
import org.apache.log4j.Logger;

public class IdsasActorProviderImpl implements IdsasActorProvider {
    private static final Logger logger = Logger.getLogger(IdsasActorProviderImpl.class);
    // Formålet med at bruge en Supplier er, at vi kan udsætte "getSecurityContext"
    // til vi er i selve ServletRequest-tråden.
    final private Supplier<SecurityContext> securitySupplier;

    public IdsasActorProviderImpl() {
        this.securitySupplier = Security::getSecurityContext;
    }

    public IdsasActorProviderImpl(Supplier<SecurityContext> securitySupplier) {
        this.securitySupplier = securitySupplier;
    }

    @Override
    public IdsasActor getActor() throws IdsasSecurityException {
        SecurityContext securityContext = securitySupplier.get();

        // Der skal være en sikkerhedsbillet
        if (!securityContext.getTicket().isPresent()) {
            throw new IdsasSecurityException("Sikkerhedsbilletten mangler");
        }

        // Sikkerhedsbilletten skal være valid
        if (!securityContext.getTicket().get().isValid()) {

            // Den må ikke være udløbet
            if (securityContext.getTicket().get().getValidTo().isPresent()) {
                ZonedDateTime validTo = securityContext.getTicket().get().getValidTo().get();
                if (validTo.isBefore(ZonedDateTime.now())) {
                    throw new IdsasSecurityException("Sikkerhedsbilletten er udløbet");
                }
            }

            // Den må ikke først være gyldig senere
            if (securityContext.getTicket().get().getValidFrom().isPresent()) {
                ZonedDateTime validFrom = securityContext.getTicket().get().getValidFrom().get();
                if (validFrom.isAfter(ZonedDateTime.now())) {
                    throw new IdsasSecurityException("Sikkerhedsbilletten er ikke gyldig endnu");
                }
            }

            String audience = securityContext.getTicket().get().getAudience().isPresent() ?
                    securityContext.getTicket().get().getAudience().get() : "";
            String created = securityContext.getTicket().get().getCreated().isPresent() ?
                    securityContext.getTicket().get().getCreated().get().toString() : "";

            throw new IdsasSecurityException("Sikkerhedsbilletten er ikke valid. audience = " +
                    audience + ", created = " + created);
        }

        // Gennemløber de enkelte brugertyper
        List<IdsasActor> matchedActors = new LinkedList<>();
        IdsasActor system = matchSystemActor(securityContext);
        if (system != null) {
            matchedActors.add(system);
        }
        IdsasActor sundhedsfaglig = matchSundhedsfagligActor(securityContext);
        if (sundhedsfaglig != null) {
            matchedActors.add(sundhedsfaglig);
        }

        if (matchedActors.size() == 0) {
            throw new IdsasSecurityException(ErrorMessages.invalidUserType());
        } else if (matchedActors.size() > 1) {
            throw new IdsasSecurityException(ErrorMessages.userTypeAmbiguous());
        }
        return matchedActors.get(0);
    }

    private IdsasActor matchSystemActor(SecurityContext securityContext) {
        if (securityContext.getActingUser().isPresent()) {
            logger.info("Systembruger kan ikke være en acting user");
            return null;
        };

        if (securityContext.getPrincipalUser().isPresent()) {
            logger.info("Systembruger kan ikke være en principal user");
            return null;
        }

        if (!securityContext.getOrganisation().isPresent()) {
            logger.info("Systembruger skal have en organisation");
            return null;
        }

        if (!securityContext.getOrganisation().get().getIdentifierFormat().equals(
                SecurityContext.Organisation.OrganisationIdentifierFormat.CVR)) {
            logger.info("Systembruger skal have en organisation af typen CVR");
            return null;
        }

        String organisationIdentifier = securityContext.getOrganisation().get().getIdentifier();
        if (organisationIdentifier.isEmpty()) {
            logger.info("Systembruger skal have en organisation med en identifier");
            return null;
        }

        return IdsasActor.createSystem(organisationIdentifier);
    }

    private IdsasActor matchSundhedsfagligActor(SecurityContext securityContext) {
        if (securityContext.getPrincipalUser().isPresent()) {
            logger.info("Sundhedsfaglig kan ikke være en principal user");
            return null;
        }

        if (!securityContext.getActingUser().isPresent()) {
            logger.info("Sundhedsfaglig skal have en acting user");
            return null;
        }

        if (!securityContext.getActingUser().get().getUserType().equals(
                SecurityContext.User.UserType.HealthcareProfessional)) {
            logger.info("Sundhedsfaglig ActingUser skal have typen HealthcareProfessional");
            return null;
        }

        // ActingUser skal have en identifier
        if (securityContext.getActingUser().get().getIdentifier().isEmpty()) {
            logger.info("Sundhedsfaglig ActingUser skal have en identifier");
            return null;
        }

        // Identifier skal have format CPR
        if (!securityContext.getActingUser().get().getIdentifierFormat().equals(
                SecurityContext.User.PersonIdentifierFormat.CPR)) {
            logger.info("Sundhedsfaglig ActingUser skal have en identifier af typen CPR");
            return null;
        }

        // Sundhedsfaglig skal have en organisation
        if (!securityContext.getOrganisation().isPresent()) {
            logger.info("Sundhedsfaglig skal have en organisation");
            return null;
        }

        // Sundhedsfaglig skal have en organisation af typen CVR
        if (!securityContext.getOrganisation().get().getIdentifierFormat().equals(
                SecurityContext.Organisation.OrganisationIdentifierFormat.CVR)) {
            logger.info("Sundhedsfaglig skal have en organisation af typen CVR");
            return null;
        }

        // Sundhedsfaglig skal have en organisation med en identifier
        String organisationIdentifier = securityContext.getOrganisation().get().getIdentifier();
        if (organisationIdentifier.isEmpty()) {
            logger.info("Sundhedsfaglig skal have en organisation med en identifier");
            return null;
        }

        return IdsasActor.createSundhedsFaglig(organisationIdentifier);
    }
}
