package dk.nsp.idsas.ws._20230601;

import javax.jws.WebService;
import javax.servlet.annotation.WebServlet;
import javax.xml.ws.BindingType;
import javax.xml.ws.soap.SOAPFaultException;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;

import dk.sundhedsdatastyrelsen.identitetssloering._2023._06._01.*;

import dk.nsp.idsas.integrations.PersonInformationClient;
import dk.nsp.idsas.integrations.SoresClient;
import dk.nsp.idsas.security.model.IdsasActor;
import dk.nsp.idsas.util.SoapFaultGenerator;
import dk.nsp.idsas.util.WhitelistType;
import dk.nsp.idsas.dao.WhitelistDao;
import dk.nsp.idsas.ws.IdsasAbstractWS;
import dk.nsp.idsas.service.BlurringService;
import dk.nsp.idsas.util.ErrorMessages;
import dk.nsp.idsas.util.AuditLogger;
import dk.sds.nsp.audit.Audit;

import static javax.xml.ws.soap.SOAPBinding.SOAP11HTTP_BINDING;

@WebServlet(name="IdsasService", urlPatterns={"/20230601/service"})
@WebService(serviceName = "IdsasService", wsdlLocation = "wsdl/20230601/idsas-service.wsdl",
        targetNamespace = "http://sundhedsdatastyrelsen.dk/identitetssloering/2023/06/01/",
        portName = "IdsasServicePort")
@BindingType(value = SOAP11HTTP_BINDING)
public class IdsasServiceWS extends IdsasAbstractWS implements IdsasServicePortType {
    private static final Logger logger = Logger.getLogger(IdsasServiceWS.class);

    @Autowired
    private BlurringService blurringService;

    @Autowired
    private AuditLogger auditLogger;

    @Override
    public String createBlurring(CreateBlurringRequestType createBlurringRequest) {

        try {
            IdsasActor actor = getValidatedIdsasActor(WhitelistType.BLURRING);
            logger.info(String.format("Enter createBlurring for cvr '%s', patient '%s'",
                    actor.getCvr(),
                    createBlurringRequest.getPatientId().getId()));

            // Tjek at CVR er whitelisted til at kalde createBlurring
            WhitelistDao whitelistDao = new WhitelistDao(dataSource);
            boolean isWhiteListed = whitelistDao.isWhitelisted(actor.getCvr(), WhitelistType.BLURRING);
            if (!isWhiteListed) {
                throw SoapFaultGenerator.generateSOAPFault(true,
                        ErrorMessages.whitelistError(actor.getCvr(), WhitelistType.BLURRING));
            }

            // Tjek at CVR mapper til en SOR kode (via SORES)
            SoresClient soresClient = new SoresClient();
            boolean cvrMapsToSor = soresClient.cvrMapsToSor(actor.getCvr());
            if (!cvrMapsToSor) {
                throw SoapFaultGenerator.generateSOAPFault(true,
                        ErrorMessages.cvrDoesNotMapToSorError(actor.getCvr()));
            }

            // Tjek at patientId har typen CPR; vi håndterer ikke andre typer lige nu
            String idClassification = createBlurringRequest.getPatientId().getClassification() == null ?
                    "null" : createBlurringRequest.getPatientId().getClassification().value();
            if (!idClassification.equals(PatientIdClassificationType.CPR.value())) {
                throw SoapFaultGenerator.generateSOAPFault(true,
                        ErrorMessages.idTypeNotSupportedError(idClassification));
            }

            // Tjek at patientId er gyldig (via "personinformation")
            String cpr = createBlurringRequest.getPatientId().getId();
            PersonInformationClient personInformationClient = new PersonInformationClient();
            boolean patientIdIsValid = personInformationClient.cprIsValid(cpr);

            if (!patientIdIsValid) {
                throw SoapFaultGenerator.generateSOAPFault(true,
                        ErrorMessages.cprIsNotValidError(cpr));
            }

            // Opretter slørring
            blurringService.createBlurring(createBlurringRequest, actor);

            // Audit logning
            auditLogger.logCreateBlurring(Audit.createAuditBuilder(), cpr, actor.getCvr());

        } catch (SOAPFaultException e) {
            throw e; // Den fanges af platformen
        } catch (Exception e) {
            logger.error("Fejl ved oprettelse af sløring", e);
            throw SoapFaultGenerator.generateSOAPFault(false, "Internal server error: " + e.getMessage());
        }

        return "OK";
    }

    @Override
    public String getCurrentSalt(EmptyRequest getCurrentSaltRequest) {
        IdsasActor actor = getValidatedIdsasActor(WhitelistType.SALT);
        logger.info(String.format("Enter get current salt for cvr '%s'", actor.getCvr()));

        auditLogger.logGetCurrentSalt(Audit.createAuditBuilder(), actor.getCvr());

        return "getCurrentSalt";
    }
}
