package dk.nsp.idsas.ws;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;

import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.annotation.DirtiesContext;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.SpringRunner;

import dk.sundhedsdatastyrelsen.identitetssloering._2023._06._01.*;

import dk.nsp.idsas.setup.DataSourceTestConfiguration;
import dk.nsp.idsas.security.model.IdsasActor;
import dk.nsp.idsas.service.BlurringService;
import dk.nsp.idsas.setup.BlurringTestConfiguration;
import dk.nsp.idsas.service.LookupService;
import dk.nsp.idsas.setup.LookupTestConfiguration;

import static org.junit.Assert.assertEquals;

@RunWith(SpringRunner.class)
@ContextConfiguration(classes = {
        DataSourceTestConfiguration.class,
        BlurringTestConfiguration.class,
        LookupTestConfiguration.class})
@DirtiesContext(classMode = DirtiesContext.ClassMode.BEFORE_EACH_TEST_METHOD)
public class IdsasServiceWSTest {

    @Autowired
    BlurringService blurringService;

    @Autowired
    LookupService lookupService;

    @Test
    public void createOneBlurring() {
        // Arrange
        String patientId = "0101019999";
        String patientIdType = "CPR";
        String organisationId = "94847464";
        LocalDateTime localDateTime = LocalDateTime.now().plusDays(1);

        // Act
        blurringService.createBlurring(createBlurringRequestType(patientId, localDateTime),
                IdsasActor.createSundhedsFaglig(organisationId));

        // Assert
        BlurredOrganisationsResponseType response = lookupService.getActiveBlurrings(patientId, patientIdType);
        List<OrganisationIdType> blurrings = response.getOrganisationIds();

        assertEquals(1, blurrings.size());
    }

    @Test
    public void createThreeBlurringsForSamePersonAndSameOrg() {
        // Arrange
        String patientId = "0101019999";
        String patientIdType = "CPR";
        String organisationId = "94847464";
        LocalDateTime localDateTime = LocalDateTime.now().plusDays(1);

        // Act
        blurringService.createBlurring(createBlurringRequestType(patientId, localDateTime),
                IdsasActor.createSundhedsFaglig(organisationId));
        blurringService.createBlurring(createBlurringRequestType(patientId, localDateTime),
                IdsasActor.createSundhedsFaglig(organisationId));
        blurringService.createBlurring(createBlurringRequestType(patientId, localDateTime),
                IdsasActor.createSundhedsFaglig(organisationId));

        // Assert
        BlurredOrganisationsResponseType response = lookupService.getActiveBlurrings(patientId, patientIdType);
        List<OrganisationIdType> blurrings = response.getOrganisationIds();

        // Kun seneste blurring er aktiv
        assertEquals(1, blurrings.size());
    }

    @Test
    public void createThreeBlurringsForSamePersonDifferentOrgs() {
        // Arrange
        String patientId = "0101019999";
        String patientIdType = "CPR";
        LocalDateTime localDateTime = LocalDateTime.now().plusDays(1);

        // Act
        blurringService.createBlurring(createBlurringRequestType(patientId, localDateTime),
                IdsasActor.createSundhedsFaglig("94847464"));
        blurringService.createBlurring(createBlurringRequestType(patientId, localDateTime),
                IdsasActor.createSundhedsFaglig("94847465"));
        blurringService.createBlurring(createBlurringRequestType(patientId, localDateTime),
                IdsasActor.createSundhedsFaglig("94847466"));

        // Assert
        BlurredOrganisationsResponseType response = lookupService.getActiveBlurrings(patientId, patientIdType);
        List<OrganisationIdType> blurrings = response.getOrganisationIds();

        assertEquals(3, blurrings.size());
    }

    @Test
    public void createBlurringInTheFarFuture() {
        // Arrange
        String patientId = "0101019999";
        String patientIdType = "CPR";
        String organisationId = "94847464";
        LocalDateTime localDateTime = LocalDateTime.now().plusYears(5);

        // Act
        blurringService.createBlurring(createBlurringRequestType(patientId, localDateTime),
                IdsasActor.createSundhedsFaglig(organisationId));

        // Assert
        BlurredOrganisationsResponseType response = lookupService.getActiveBlurrings(patientId, patientIdType);
        List<OrganisationIdType> blurrings = response.getOrganisationIds();

        assertEquals(1, blurrings.size());
    }

    @Test
    public void pastBlurringsAreNotActive() {
        // Arrange
        String patientId = "0101019999";
        String patientIdType = "CPR";
        String organisationId = "94847464";
        LocalDateTime localDateTime = LocalDateTime.now().minusYears(5);

        // Act
        blurringService.createBlurring(createBlurringRequestType(patientId, localDateTime),
                IdsasActor.createSundhedsFaglig(organisationId));

        // Assert
        BlurredOrganisationsResponseType response = lookupService.getActiveBlurrings(patientId, patientIdType);
        List<OrganisationIdType> blurrings = response.getOrganisationIds();
        assertEquals(0, blurrings.size());
    }


    private CreateBlurringRequestType createBlurringRequestType(String inputId, LocalDateTime datetime) {
        return new CreateBlurringRequestType() {{
            setPatientId(new PatientIdType() {{
                setId(inputId);
                setClassification(PatientIdClassificationType.CPR);
            }});
            setEndDateTime(toXmlGregorian(datetime));
        }};
    }

    private XMLGregorianCalendar toXmlGregorian(LocalDateTime datetime) {
        try {
            GregorianCalendar gregorianCalendar = (GregorianCalendar) GregorianCalendar.getInstance();
            gregorianCalendar.setTime(Date.from(datetime.toInstant(ZoneOffset.UTC)));
            return DatatypeFactory.newInstance().newXMLGregorianCalendar(gregorianCalendar);
        } catch (DatatypeConfigurationException e) {
            throw new RuntimeException(e);
        }
    }

}
