/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsp.idsas.personinformation;


import com.fasterxml.jackson.databind.ObjectMapper;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.log4j.Logger;

import java.io.IOException;
import java.time.Duration;
import java.util.Map;

public class PersonInformationClient {

    private static final String APPLICATION_JSON = "application/json";

    private static final String CONTENT_TYPE = "Content-Type";

    private static final Logger LOGGER = Logger.getLogger(PersonInformationClient.class);

    private final ObjectMapper objectMapper = new ObjectMapper();

    private final ServiceErrorCallback serviceErrorCallback;

    private final HttpClient httpClient;

    private final String url;

    public PersonInformationClient(HttpClient httpClient, String url, ServiceErrorCallback serviceErrorCallback) {
        this.httpClient = httpClient;
        this.url = url;
        this.serviceErrorCallback = serviceErrorCallback;
    }

    public PersonInformationClient(HttpClient httpClient, String url, Duration errorCountDuration) {
        this(httpClient, url, new ServiceErrorCallback(errorCountDuration));
    }

    public boolean personExists(String cpr) throws PersonInformationClientException {
        HttpUriRequest cprRequest = buildGetPersonInformationRequest(cpr);
        return invokeService(cprRequest);
    }

    public int getErrorCount() {
        return serviceErrorCallback.getErrorCount();
    }

    private HttpUriRequest buildGetPersonInformationRequest(String cpr) {
        // Example request string: http://localhost:8086/stamdata-personinformation/person/1212121212
        return new HttpGet(String.format("%s/person/%s", url, cpr));
    }

    private boolean invokeService(HttpUriRequest request) throws PersonInformationClientException {
        try {
            // Using a ResponseHandler makes the httpclient framework handle resource allocation/deallocation for us.
            // If you decide to not use a ResponseHandler anymore, you must ensure that resources (connections) are properly released!
            boolean response = httpClient.execute(request, this::handleResponseWithExceptionWrapper);
            serviceErrorCallback.registerServiceSuccess();
            return response;
        } catch (IOException e) {
            String errorMessage = "CPR-nummeret kunne ikke valideres, fordi kaldet til PersonInformation-servicen gik galt.";
            LOGGER.error(errorMessage, e);
            serviceErrorCallback.registerServiceError();
            throw new PersonInformationClientException(errorMessage, e);
        } catch (PersonInformationClientExceptionWrapper e) {
            serviceErrorCallback.registerServiceError();
            throw e.getPersonInformationClientException();
        }
    }

    private boolean handleResponseWithExceptionWrapper(HttpResponse response) {
        try {
            return handleResponse(response);
        } catch (PersonInformationClientException e) {
            throw new PersonInformationClientExceptionWrapper(e);
        }
    }

    private boolean handleResponse(HttpResponse response) throws PersonInformationClientException {
        int statusCode = response.getStatusLine().getStatusCode();
        switch (statusCode) {
            case HttpStatus.SC_OK: {
                boolean contentTypeOk = contentTypeOk(response);
                Map<String, Object> responseJson = parseResponseAsJson(response);
                if (!contentTypeOk || !responseJson.containsKey("cpr")) {
                    LOGGER.error("Fejl i retursvar fra PersonInformation (200)");
                    throw new PersonInformationClientException("Fejl i retursvar fra PersonInformation");
                }
                return true;
            }
            case HttpStatus.SC_NOT_FOUND: {
                boolean contentTypeOk = contentTypeOk(response);
                if (!contentTypeOk) {
                    LOGGER.error("Fejl i retursvar fra PersonInformation (404) forkert contenttype");
                    throw new PersonInformationClientException("Fejl i retursvar fra PersonInformation");
                }
                parseResponseAsJson(response);
                return false;
            }
            default:
                throw new PersonInformationClientException(String.format("CPR-nummeret kunne ikke valideres, fordi kaldet til PersonInformation-servicen returnerede en uventet statuskode: %d", statusCode));
        }
    }


    private Map<String, Object> parseResponseAsJson(HttpResponse response) throws PersonInformationClientException {
        try {
            return objectMapper.readValue(response.getEntity().getContent(), Map.class);
        } catch (UnsupportedOperationException | IOException e) {
            LOGGER.error("Kunne ikke parse response body til json", e);
            throw new PersonInformationClientException("Fejl i fortolkning af svar fra PersonInformation", e);
        }
    }

    private boolean contentTypeOk(HttpResponse response) {
        Header headers = response.getFirstHeader(CONTENT_TYPE);
        return headers != null && headers.getValue().equalsIgnoreCase(APPLICATION_JSON);
    }

    private class PersonInformationClientExceptionWrapper extends RuntimeException {
        private final PersonInformationClientException personInformationClientException;

        public PersonInformationClientExceptionWrapper(PersonInformationClientException personInformationClientException) {
            this.personInformationClientException = personInformationClientException;
        }

        public PersonInformationClientException getPersonInformationClientException() {
            return personInformationClientException;
        }
    }

}
