package dk.nsp.idsas.personinformation;

import org.apache.http.HttpResponse;
import org.apache.http.StatusLine;
import org.apache.http.client.HttpClient;
import org.apache.http.client.ResponseHandler;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.entity.BasicHttpEntity;
import org.apache.http.message.BasicHttpResponse;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;
import org.mockito.stubbing.Answer;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;

import static org.junit.Assert.*;

public class PersonInformationClientTest {

    private HttpClient httpClient;
    private final String url = "test-url";
    private ServiceErrorCallback serviceErrorCallback;
    private PersonInformationClient personInformationClient;

    @Before
    public void setUp() {
        httpClient = Mockito.mock(HttpClient.class);
        serviceErrorCallback = Mockito.mock(ServiceErrorCallback.class);
        personInformationClient = new PersonInformationClient(httpClient, url, serviceErrorCallback);
    }

    @Test
    public void personExists_WhenServiceReturnsValidResponse_ReturnsTrueAndRegistersServiceSuccess() throws PersonInformationClientException, IOException {
        HttpResponse httpResponse = createHttpResponse(200, "application/json", "{\"cpr\": \"1234\"}");
        Mockito.doAnswer(createHandleResponseAnswer(httpResponse))
                .when(httpClient).execute(Mockito.any(HttpUriRequest.class), Mockito.any(ResponseHandler.class));
        String cpr = "1234";

        boolean personExists = personInformationClient.personExists(cpr);

        assertTrue(personExists);
        String expectedPath = url + "/person/" + cpr;
        Mockito.verify(httpClient).execute(
                Mockito.argThat(request -> expectedPath.equals(request.getURI().getPath())),
                Mockito.any(ResponseHandler.class));
        Mockito.verify(serviceErrorCallback).registerServiceSuccess();
    }

    @Test
    public void personExists_WhenClientThrowsIOException_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        Mockito.doThrow(IOException.class)
                .when(httpClient).execute(Mockito.any(HttpUriRequest.class), Mockito.any(ResponseHandler.class));

        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personExists("1234"));

        assertEquals("CPR-nummeret kunne ikke valideres, fordi kaldet til PersonInformation-servicen gik galt.", e.getMessage());
        Mockito.verify(serviceErrorCallback).registerServiceError();
    }

    @Test
    public void personExists_WhenServiceReturnsResponse200WithContentTypeText_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        HttpResponse httpResponse = createHttpResponse(200, "text/plain", "{\"cpr\": \"1234\"}");
        Mockito.doAnswer(createHandleResponseAnswer(httpResponse))
                .when(httpClient).execute(Mockito.any(HttpUriRequest.class), Mockito.any(ResponseHandler.class));

        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personExists("1234"));

        assertEquals("Fejl i retursvar fra PersonInformation", e.getMessage());
        Mockito.verify(serviceErrorCallback).registerServiceError();
    }

    @Test
    public void personExists_WhenServiceReturnsResponse200WithInvalidJson_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        HttpResponse httpResponse = createHttpResponse(200, "application/json", "invalid body");
        Mockito.doAnswer(createHandleResponseAnswer(httpResponse))
                .when(httpClient).execute(Mockito.any(HttpUriRequest.class), Mockito.any(ResponseHandler.class));

        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personExists("1234"));

        assertEquals("Fejl i fortolkning af svar fra PersonInformation", e.getMessage());
        Mockito.verify(serviceErrorCallback).registerServiceError();
    }

    @Test
    public void personExists_WhenServiceReturnsResponse200WithoutCprEntry_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        HttpResponse httpResponse = createHttpResponse(200, "application/json", "{\"test\": \"1234\"}");
        Mockito.doAnswer(createHandleResponseAnswer(httpResponse))
                .when(httpClient).execute(Mockito.any(HttpUriRequest.class), Mockito.any(ResponseHandler.class));

        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personExists("1234"));

        assertEquals("Fejl i retursvar fra PersonInformation", e.getMessage());
        Mockito.verify(serviceErrorCallback).registerServiceError();
    }

    @Test
    public void personExists_WhenServiceReturnsResponse404WithContentTypeText_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        HttpResponse httpResponse = createHttpResponse(404, "text/plain", "{\"cpr\": \"1234\"}");
        Mockito.doAnswer(createHandleResponseAnswer(httpResponse))
                .when(httpClient).execute(Mockito.any(HttpUriRequest.class), Mockito.any(ResponseHandler.class));

        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personExists("1234"));

        assertEquals("Fejl i retursvar fra PersonInformation", e.getMessage());
        Mockito.verify(serviceErrorCallback).registerServiceError();
    }

    @Test
    public void personExists_WhenServiceReturnsResponse404WithInvalidJson_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        HttpResponse httpResponse = createHttpResponse(404, "application/json", "invalid body");
        Mockito.doAnswer(createHandleResponseAnswer(httpResponse))
                .when(httpClient).execute(Mockito.any(HttpUriRequest.class), Mockito.any(ResponseHandler.class));

        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personExists("1234"));

        assertEquals("Fejl i fortolkning af svar fra PersonInformation", e.getMessage());
        Mockito.verify(serviceErrorCallback).registerServiceError();
    }

    @Test
    public void personExists_WhenServiceReturnsResponse404WithValidJson_ReturnsFalseAndRegistersServiceSuccess() throws PersonInformationClientException, IOException {
        HttpResponse httpResponse = createHttpResponse(404, "application/json", "{\"test\": \"test\"}");
        Mockito.doAnswer(createHandleResponseAnswer(httpResponse))
                .when(httpClient).execute(Mockito.any(HttpUriRequest.class), Mockito.any(ResponseHandler.class));

        boolean personExists = personInformationClient.personExists("1234");

        assertFalse(personExists);
        Mockito.verify(serviceErrorCallback).registerServiceSuccess();
    }

    @Test
    public void personExists_WhenServiceReturnsResponse400_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        HttpResponse httpResponse = createHttpResponse(400, "application/json", "{\"test\": \"test\"}");
        Mockito.doAnswer(createHandleResponseAnswer(httpResponse))
                .when(httpClient).execute(Mockito.any(HttpUriRequest.class), Mockito.any(ResponseHandler.class));

        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personExists("1234"));

        assertEquals("CPR-nummeret kunne ikke valideres, fordi kaldet til PersonInformation-servicen returnerede en uventet statuskode: 400", e.getMessage());
        Mockito.verify(serviceErrorCallback).registerServiceError();
    }

    private HttpResponse createHttpResponse(int statusCode, String contentType, String body) {
        StatusLine statusLine = Mockito.mock(StatusLine.class);
        Mockito.when(statusLine.getStatusCode()).thenReturn(statusCode);
        BasicHttpResponse httpResponse = new BasicHttpResponse(statusLine);
        BasicHttpEntity httpEntity = new BasicHttpEntity();
        httpEntity.setContent(inputStreamFromString(body));
        httpResponse.setEntity(httpEntity);
        httpResponse.setHeader("Content-Type", contentType);
        return httpResponse;
    }

    @SuppressWarnings("unchecked")
    private Answer<Boolean> createHandleResponseAnswer(HttpResponse response) {
        return arg -> ((ResponseHandler<Boolean>) arg.getArgument(1)).handleResponse(response);
    }

    private InputStream inputStreamFromString(String input) {
        return new ByteArrayInputStream(input.getBytes(StandardCharsets.UTF_8));
    }
}